@@ -85,6 +85,8 @@ G_DEFINE_TYPE_WITH_CODE (SswSheetBody, ssw_sheet_body,
 #define PRIV(l) ((SswSheetBodyPrivate *)ssw_sheet_body_get_instance_private ((SswSheetBody *)l))
 
 
+static void start_editing (SswSheetBody *body, GdkEvent *e);
+
 static void
 limit_selection (SswSheetBody *body)
 {
@@ -104,7 +106,7 @@ limit_selection (SswSheetBody *body)
 }
 
 
-static void
+void
 normalise_selection (const SswRange *in, SswRange *out)
 {
   if (in->start_y < in->end_y)
@@ -259,36 +261,39 @@ draw_selection (SswSheetBody *body, cairo_t *cr)
       ||
       (priv->selection->start_y != priv->selection->end_y))
     {
-
       SswRange mySelect;
       normalise_selection (priv->selection, &mySelect);
-
-      gint xpos_start, ypos_start;
-      if (0 != ssw_sheet_axis_find_boundary (priv->haxis, mySelect.start_x,
-					     &xpos_start, NULL))
+      gint xpos_start = 0;
+      gint ypos_start = 0;
+      if (0 < ssw_sheet_axis_find_boundary (priv->haxis, mySelect.start_x,
+					    &xpos_start,  NULL))
+	goto done;
+      
+      if (0 < ssw_sheet_axis_find_boundary (priv->vaxis, mySelect.start_y,
+					    &ypos_start,  NULL))
 	goto done;
 
-      if (0 != ssw_sheet_axis_find_boundary (priv->vaxis, mySelect.start_y,
-					     &ypos_start, NULL))
-	goto done;
-
-      gint xlimit = width;
-      gint ylimit = height;
       gint xpos_end, ypos_end, yextent, xextent;
-      if (0 == ssw_sheet_axis_find_boundary (priv->haxis,
-					     mySelect.end_x, &xpos_end, &xextent))
-	xlimit = xpos_end + xextent;
-
-
-      if (0 == ssw_sheet_axis_find_boundary (priv->vaxis,
-					     mySelect.end_y, &ypos_end, &yextent))
-	ylimit = ypos_end + yextent;
+      gint xrel = ssw_sheet_axis_find_boundary (priv->haxis,
+					       mySelect.end_x, &xpos_end,
+					       &xextent);
+      if (xrel < 0)
+	goto done;
+      
+      gint xsize = (xrel == 0) ? xpos_end - xpos_start + xextent : width;
+      
 
+      gint yrel = ssw_sheet_axis_find_boundary (priv->vaxis,
+						mySelect.end_y, &ypos_end,
+						&yextent);
+      if (yrel < 0)
+	goto done;
+      
+      gint ysize = (yrel == 0) ? ypos_end - ypos_start + yextent : height;
 
       gtk_render_background (sc, cr,
 			     xpos_start, ypos_start,
-			     xlimit - xpos_start,
-			     ylimit - ypos_start);
+			     xsize, ysize);
     }
 
  done:
@@ -395,6 +400,12 @@ __draw (GtkWidget *widget, cairo_t *cr)
 
   GtkCssProvider *cp = gtk_css_provider_new ();
 
+  gint yy = ssw_sheet_axis_find_boundary (priv->vaxis, active_row, NULL, NULL);
+  gint xx = ssw_sheet_axis_find_boundary (priv->haxis, active_col, NULL, NULL);
+
+  if (yy == 0 && xx == 0 && priv->editor == NULL)
+    start_editing (body, NULL);
+  
   if ((gtk_widget_is_focus (widget) ||
        (priv->editor && gtk_widget_is_focus (priv->editor))))
     gtk_css_provider_load_from_data (cp, focused_border, strlen (focused_border), 0);
@@ -608,7 +619,6 @@ set_selection (SswSheetBody *body, gint start_x, gint start_y, gint end_x, gint
   priv->selection->end_x = end_x;
   priv->selection->end_y = end_y;
 
-
   announce_selection (body);
 }
 
@@ -1119,13 +1129,13 @@ select_entire_row (SswSheetBody *body, gint i, guint state, gpointer ud)
   set_selection (body,
 		 0, start_y,
 		 ssw_sheet_axis_get_size (priv->haxis) - 1, i);
+  start_editing (body, NULL);
 }
 
 static void
 select_entire_column (SswSheetBody *body, gint i, guint state, gpointer ud)
 {
   PRIV_DECL (body);
-
   GdkDisplay *disp = gtk_widget_get_display (GTK_WIDGET (body));
   GdkKeymap *km = gdk_keymap_get_for_display (disp);
   GdkModifierType extend_mask =
@@ -1136,6 +1146,7 @@ select_entire_column (SswSheetBody *body, gint i, guint state, gpointer ud)
   set_selection (body,
 		 start_x, 0,
 		 i, ssw_sheet_axis_get_size (priv->vaxis) - 1);
+  start_editing (body, NULL);
 }
 
 
@@ -1264,7 +1275,8 @@ __set_property (GObject *object,
       if (priv->data_model)
 	g_object_unref (priv->data_model);
       priv->data_model = g_value_get_object (value);
-      g_signal_connect (priv->data_model, "items-changed", G_CALLBACK (on_data_change), body);
+      g_signal_connect_object (priv->data_model, "items-changed",
+			       G_CALLBACK (on_data_change), body, 0);
       g_object_ref (priv->data_model);
       break;
     default:
@@ -1674,7 +1686,7 @@ append_value_to_string (SswSheetBody *body,
 
 
 static void
-clipit_html (SswSheetBody *body, GString *output)
+clipit_html (SswSheetBody *body, GString *output, SswRange *source_range)
 {
   PRIV_DECL (body);
   if (!priv->haxis || !priv->vaxis || !priv->data_model)
@@ -1685,14 +1697,12 @@ clipit_html (SswSheetBody *body, GString *output)
 
   gint row;
   gint col;
-  SswRange mySelection;
-  normalise_selection (priv->selection, &mySelection);
-  for (row = mySelection.start_y ; row <= mySelection.end_y; ++row)
+  for (row = source_range->start_y ; row <= source_range->end_y; ++row)
     {
       GtkTreeIter iter;
       gtk_tree_model_iter_nth_child (priv->data_model, &iter, NULL, row);
       g_string_append (output, "<tr>\n");
-      for (col = mySelection.start_x; col <= mySelection.end_x; ++col)
+      for (col = source_range->start_x; col <= source_range->end_x; ++col)
 	{
 	  if (row < gtk_tree_model_iter_n_children (priv->data_model, NULL)
 	      && col < gtk_tree_model_get_n_columns (priv->data_model))
@@ -1711,7 +1721,7 @@ clipit_html (SswSheetBody *body, GString *output)
 
 
 static void
-clipit_ascii (SswSheetBody *body, GString *output)
+clipit_ascii (SswSheetBody *body, GString *output, SswRange *source_range)
 {
   PRIV_DECL (body);
   if (!priv->haxis || !priv->vaxis || !priv->data_model)
@@ -1719,14 +1729,12 @@ clipit_ascii (SswSheetBody *body, GString *output)
 
   gint row;
   gint col;
-  SswRange mySelection;
-  normalise_selection (priv->selection, &mySelection);
-  for (row = mySelection.start_y ; row <= mySelection.end_y; ++row)
+  for (row = source_range->start_y ; row <= source_range->end_y; ++row)
     {
       GtkTreeIter iter;
       gtk_tree_model_iter_nth_child (priv->data_model, &iter, NULL, row);
 
-      for (col = mySelection.start_x; col <= mySelection.end_x; ++col)
+      for (col = source_range->start_x; col <= source_range->end_x; ++col)
 	{
 	  if (priv->data_model
 	      && row < gtk_tree_model_iter_n_children (priv->data_model, NULL)
@@ -1734,7 +1742,7 @@ clipit_ascii (SswSheetBody *body, GString *output)
 	    {
 	      append_value_to_string (body, &iter, col, row, output);
 
-	      if (col < mySelection.end_x)
+	      if (col < source_range->end_x)
 		g_string_append (output, "\t");
 	    }
 	}
@@ -1743,7 +1751,7 @@ clipit_ascii (SswSheetBody *body, GString *output)
 }
 
 static void
-clipit_utf8 (SswSheetBody *body, GString *output)
+clipit_utf8 (SswSheetBody *body, GString *output, SswRange *source_range)
 {
   PRIV_DECL (body);
   if (!priv->haxis || !priv->vaxis || !priv->data_model)
@@ -1751,22 +1759,20 @@ clipit_utf8 (SswSheetBody *body, GString *output)
 
   gint row;
   gint col;
-  SswRange mySelection;
-  normalise_selection (priv->selection, &mySelection);
-  for (row = mySelection.start_y ; row <= mySelection.end_y; ++row)
+  for (row = source_range->start_y ; row <= source_range->end_y; ++row)
     {
       GtkTreeIter iter;
       gtk_tree_model_iter_nth_child (priv->data_model, &iter, NULL, row);
 
-      for (col = mySelection.start_x; col <= mySelection.end_x; ++col)
+      for (col = source_range->start_x; col <= source_range->end_x; ++col)
 	{
 	  if (row < gtk_tree_model_iter_n_children (priv->data_model, NULL)
 	      && col < gtk_tree_model_get_n_columns (priv->data_model))
 	    {
 	      append_value_to_string (body, &iter, col, row, output);
 
-	      if (col < mySelection.end_x)
-		g_string_insert_unichar (output, -1, 0x2502);
+	      if (col < source_range->end_x)
+		g_string_append_c (output, '\t');
 	    }
 	}
       g_string_append (output, "\n");
@@ -1788,17 +1794,19 @@ get_func (GtkClipboard *clipboard,
       return;
     }
 
+  SswRange *source_range = g_object_get_data (G_OBJECT (clipboard), "source-range");
+
   GString *stuff = g_string_new ("");
   switch (info)
     {
     case TARGET_STRING:
-      clipit_ascii (body, stuff);
+      clipit_ascii (body, stuff, source_range);
       break;
     case TARGET_UTF8_STRING:
-      clipit_utf8 (body, stuff);
+      clipit_utf8 (body, stuff, source_range);
       break;
     case TARGET_HTML:
-      clipit_html (body, stuff);
+      clipit_html (body, stuff, source_range);
       break;
     default:
       g_print ("Request for unknown target %d\n", info);
@@ -2212,8 +2220,6 @@ on_remove_widget (GtkCellEditable *e, gpointer ud)
   SswSheetBody *body = ud;
   PRIV_DECL (body);
 
-  g_signal_handlers_disconnect_by_func (e, on_remove_widget, ud);
-
   gtk_widget_destroy (GTK_WIDGET (e));
   priv->editor = NULL;
   gtk_widget_hide (priv->active_cell_holder);
@@ -2225,6 +2231,15 @@ on_changed (GtkEditable *editable, gpointer user_data)
   g_object_set (editable, "editing-canceled", FALSE, NULL);
 }
 
+static gboolean
+on_focus_out (GtkWidget *w, GdkEvent *e, gpointer ud)
+{
+  /* It seems that GtkEntry by default emits the "widget-remove" signal,
+     in response to focus-out.  We don't want that to happen, so disable
+     it here. */
+  return TRUE;
+}
+
 static void
 text_editing_started (GtkCellRenderer *cell,
                       GtkCellEditable *editable,
@@ -2250,14 +2265,6 @@ text_editing_started (GtkCellRenderer *cell,
   if (! gtk_widget_is_visible (GTK_WIDGET (body)))
     return;
 
-  gint vlocation, vsize;
-  gint hlocation, hsize;
-  if (0 != ssw_sheet_axis_find_boundary (priv->vaxis, row, &vlocation, &vsize))
-    return;
-
-  if (0 != ssw_sheet_axis_find_boundary (priv->haxis, col, &hlocation, &hsize))
-    return;
-
   if (priv->editor)
     {
       gtk_cell_renderer_stop_editing (GTK_CELL_RENDERER (cell),
@@ -2273,6 +2280,15 @@ text_editing_started (GtkCellRenderer *cell,
 					&color);
 #endif
 
+  priv->editor = NULL;
+
+  gint vlocation, vsize;
+  gint hlocation, hsize;
+  if (0 != ssw_sheet_axis_find_boundary (priv->vaxis, row, &vlocation, &vsize))
+    return;
+
+  if (0 != ssw_sheet_axis_find_boundary (priv->haxis, col, &hlocation, &hsize))
+    return;
 
   priv->editor = GTK_WIDGET (editable);
   g_signal_connect (editable, "remove-widget",
@@ -2309,6 +2325,7 @@ text_editing_started (GtkCellRenderer *cell,
 
   if (GTK_IS_ENTRY (editable))
     {
+      g_signal_connect (editable, "focus-out-event", G_CALLBACK (on_focus_out), NULL);
       gtk_entry_set_text (GTK_ENTRY (editable), "");
       g_signal_connect (editable, "changed", G_CALLBACK (on_changed), NULL);
     }
