@@ -25,6 +25,7 @@
 #include "ssw-sheet-body.h"
 #include "ssw-marshaller.h"
 #include "ssw-xpaned.h"
+#include "ssw-paste.h"
 
 #define P_(X) (X)
 
@@ -73,6 +74,8 @@ enum
     PROP_SPLIT,
     PROP_GRIDLINES,
     PROP_EDITABLE,
+    PROP_HORIZONTAL_DRAGGABLE,
+    PROP_VERTICAL_DRAGGABLE,
     PROP_RENDERER_FUNC,
     PROP_RENDERER_FUNC_DATUM,
     PROP_CONVERT_FWD_FUNC,
@@ -184,19 +187,19 @@ __dispose (GObject *object)
 
 
 static void
-resize_vmodel (GtkTreeModel *tm, guint posn, guint rm, guint add, SswSheet *sheet)
+resize_vmodel (GtkTreeModel *tm, guint posn, guint rm, guint add, GListModel *vmodel)
 {
   gint rows = gtk_tree_model_iter_n_children (tm, NULL);
 
-  g_object_set (sheet->vmodel, "size", rows, NULL);
+  g_object_set (vmodel, "size", rows, NULL);
 }
 
 static void
-resize_hmodel (GtkTreeModel *tm, guint posn, guint rm, guint add, SswSheet *sheet)
+resize_hmodel (GtkTreeModel *tm, guint posn, guint rm, guint add, GListModel *hmodel)
 {
   gint cols = gtk_tree_model_get_n_columns (tm);
 
-  g_object_set (sheet->hmodel, "size", cols, NULL);
+  g_object_set (hmodel, "size", cols, NULL);
 }
 
 
@@ -253,6 +256,18 @@ __set_property (GObject *object,
       }
       break;
 
+    case PROP_HORIZONTAL_DRAGGABLE:
+	for (i = 0; i < DIM; ++i)
+	  g_object_set (SSW_SHEET_AXIS (sheet->horizontal_axis[i]),
+			"draggable",  g_value_get_boolean (value), NULL);
+	break;
+
+    case PROP_VERTICAL_DRAGGABLE:
+	for (i = 0; i < DIM; ++i)
+	  g_object_set (SSW_SHEET_AXIS (sheet->vertical_axis[i]),
+			"draggable",  g_value_get_boolean (value), NULL);
+	break;
+
     case PROP_CONVERT_FWD_FUNC:
       {
 	gpointer p = g_value_get_pointer (value);
@@ -359,7 +374,7 @@ __set_property (GObject *object,
 	  int n_rows = gtk_tree_model_iter_n_children (GTK_TREE_MODEL (sheet->data_model), NULL);
 	  g_object_set (sheet->vmodel, "size", n_rows, NULL);
 	  g_signal_connect_object (sheet->data_model, "items-changed",
-				   G_CALLBACK (resize_vmodel), sheet, 0);
+				   G_CALLBACK (resize_vmodel), sheet->vmodel, 0);
 	}
 
       if (SSW_IS_AXIS_MODEL (sheet->hmodel))
@@ -367,7 +382,7 @@ __set_property (GObject *object,
 	  int n_cols = gtk_tree_model_get_n_columns (GTK_TREE_MODEL (sheet->data_model));
 	  g_object_set (sheet->hmodel, "size", n_cols, NULL);
 	  g_signal_connect_object (sheet->data_model, "items-changed",
-				   G_CALLBACK (resize_hmodel), sheet, 0);
+				   G_CALLBACK (resize_hmodel), sheet->hmodel, 0);
 	}
 
       arrange (sheet);
@@ -462,8 +477,6 @@ ssw_sheet_class_init (SswSheetClass *class)
 			 GTK_TYPE_TREE_MODEL,
 			 G_PARAM_READWRITE);
 
-
-
   GParamSpec *split_spec =
     g_param_spec_boolean ("split",
 			  P_("Split View"),
@@ -471,6 +484,19 @@ ssw_sheet_class_init (SswSheetClass *class)
 			  FALSE,
 			  G_PARAM_READWRITE | G_PARAM_CONSTRUCT);
 
+  GParamSpec *horizontal_draggable_spec =
+    g_param_spec_boolean ("horizontal-draggable",
+			  P_("Horizontal_Draggable"),
+			  P_("If TRUE, items in the horizontal axis can be dragged."),
+			  FALSE,
+			  G_PARAM_WRITABLE | G_PARAM_CONSTRUCT);
+
+  GParamSpec *vertical_draggable_spec =
+    g_param_spec_boolean ("vertical-draggable",
+			  P_("Vertical_Draggable"),
+			  P_("If TRUE, items in the vertical axis can be dragged."),
+			  FALSE,
+			  G_PARAM_WRITABLE | G_PARAM_CONSTRUCT);
 
   GParamSpec *gridlines_spec =
     g_param_spec_boolean ("gridlines",
@@ -547,6 +573,13 @@ ssw_sheet_class_init (SswSheetClass *class)
                                    PROP_EDITABLE,
                                    editable_spec);
 
+  g_object_class_install_property (object_class,
+                                   PROP_HORIZONTAL_DRAGGABLE,
+                                   horizontal_draggable_spec);
+
+  g_object_class_install_property (object_class,
+                                   PROP_VERTICAL_DRAGGABLE,
+                                   vertical_draggable_spec);
 
   signals [ROW_HEADER_CLICKED] =
     g_signal_new ("row-header-clicked",
@@ -827,6 +860,12 @@ ssw_sheet_set_clip (SswSheet *sheet, GtkClipboard *clip)
   if (!sheet->data_model)
     return;
 
+  SswRange *source_range = g_object_get_data (G_OBJECT (clip), "source-range");
+  g_free (source_range);
+  source_range = g_malloc (sizeof (*source_range));
+  g_object_set_data (G_OBJECT (clip), "source-range", source_range);
+  normalise_selection (sheet->selection, source_range);
+  
   ssw_sheet_body_set_clip (SSW_SHEET_BODY (sheet->selected_body), clip);
 }
 
@@ -861,18 +900,8 @@ ssw_sheet_get_active_cell (SswSheet *sheet,
 
 
 
-struct paste_state
-{
-  gint col0;
-  gint row0;
-  gint col;
-  gint row;
-  SswSheet *sheet;
-  ssw_sheet_set_cell set_cell;
-};
-
-static void
-paste_datum (const gchar *x, size_t len, struct paste_state *t)
+void
+paste_insert_datum (const gchar *x, size_t len, const struct paste_state *t)
 {
   SswSheet *sheet = t->sheet;
 
@@ -890,7 +919,12 @@ paste_datum (const gchar *x, size_t len, struct paste_state *t)
   if (rcf (sheet->data_model, col, row, x, &value))
     t->set_cell (sheet->data_model, col, row, &value);
   g_value_unset (&value);
+}
 
+static void
+paste_datum (const gchar *x, size_t len, struct paste_state *t)
+{
+  paste_insert_datum (x, len, t);
   t->col ++;
 }
 
@@ -947,7 +981,7 @@ utf8_tab_delimited_parse (GtkClipboard *clip, GtkSelectionData *sd,
 {
   struct paste_state *ps = user_data;
   SswSheet *sheet = SSW_SHEET (ps->sheet);
-  const gchar *data = gtk_selection_data_get_data (sd);
+  const guchar *data = gtk_selection_data_get_data (sd);
   gint len = gtk_selection_data_get_length (sd);
 
   if (len < 0)
@@ -978,11 +1012,14 @@ target_marshaller (GtkClipboard *clip, GdkAtom *atoms, gint n_atoms,
 
   for (i = 0; i < n_atoms; ++i)
     {
-      /* Right now this is the only target we support */
-      if (atoms[i] == gdk_atom_intern ("UTF8_STRING", TRUE))
+      if (atoms[i] == gdk_atom_intern ("text/html", TRUE))
 	{
-	  gtk_clipboard_request_contents (clip, atoms[i], utf8_tab_delimited_parse,
-					  ps);
+	  gtk_clipboard_request_contents (clip, atoms[i], html_parse, ps);
+	  break;
+	}
+      else if (atoms[i] == gdk_atom_intern ("UTF8_STRING", TRUE))
+	{
+	  gtk_clipboard_request_contents (clip, atoms[i], utf8_tab_delimited_parse, ps);
 	  break;
 	}
     }
